from json import JSONDecodeError

import gspread
import google.auth
import requests
from googleapiclient.discovery import build
from df2gspread_ext import df2gspread as d2g
from df2gspread_ext import gspread2df as g2d
from gspread_pandas import Spread
from oauth2client.service_account import ServiceAccountCredentials
from retry import retry

SCOPES = ['https://spreadsheets.google.com/feeds',
          'https://www.googleapis.com/auth/drive',
          'https://www.googleapis.com/auth/spreadsheets']


class GSheetsConnector:
    def __init__(self, confs_path=None, auth_type='service_accounts', json_keyfile_dict=None, logger=None):
        self.confs_path = confs_path
        self.json_keyfile_dict = json_keyfile_dict
        self.auth_type = auth_type

        if self.json_keyfile_dict is not None or self.confs_path is not None:
            if self.json_keyfile_dict is None:
                self.creds = ServiceAccountCredentials.from_json_keyfile_name(
                    self.confs_path, scopes=SCOPES)
            else:
                self.creds = ServiceAccountCredentials.from_json_keyfile_dict(
                    self.json_keyfile_dict, scopes=SCOPES)
        else:
            self.creds, project = google.auth.default(scopes=SCOPES)

        self.service = build('sheets', 'v4', credentials=self.creds, cache_discovery=False)
        self.gspread = gspread.authorize(self.creds)

    def get_sheet_id_by_name(self, spreadsheet, worksheet_name):
        for wk in spreadsheet.worksheets():
            if wk.title == worksheet_name:
                return wk.id

    def delete_cells(self, spreadsheet_key, worksheet_name, start_index=1, end_index=30, dimension='COLUMNS',
                     additional_base_sheet=False):
        """
        :param spreadsheet_key: id for Spreadsheet taken from URL
        :param worksheet_name: name as visibile in worksheet
        :param start_index:
        :param end_index:
        :param dimension:
        :param additional_base_sheet: whethere to clean also base sheet
        :return:
        """
        spreadsheet = self.gspread.open_by_key(spreadsheet_key)
        worksheets_to_delete = [worksheet_name, 'Foglio1'] if additional_base_sheet is True else [worksheet_name]
        for wk_name in worksheets_to_delete:
            delete_body = {
                "requests": [
                    {
                        "deleteDimension": {
                            "range": {
                                "sheetId": self.get_sheet_id_by_name(spreadsheet, wk_name),
                                "dimension": dimension,
                                "startIndex": start_index,
                                "endIndex": end_index
                            }
                        }
                    }
                ]
            }
            try:
                spreadsheet.batch_update(delete_body)
            except gspread.exceptions.APIError as e:
                if "Invalid requests[0].deleteDimension: Cannot delete a column that doesn't exist." in e.response.json().get(
                        'error', {}).get('message'):
                    pass
                else:
                    raise e

    @retry((requests.exceptions.ReadTimeout, gspread.exceptions.APIError), tries=3, delay=2)
    def pd_to_gsheet(self, df, spreadsheet_key, worksheet_name, value_input_option='USER_ENTERED', clean=True, use_df2gsprad=True):
        """
        :param df: pandas DataFrame
        :param spreadsheet_key: id for Spreadsheet taken from URL
        :param worksheet_name: name as visibile in worksheet
        :param value_input_option: 'USER_ENTERED' if scope is to maintain column types from pandas DataFrame
        :param clean: whether to clean or not
        :param use_df2gsprad:
        :return:
        """
        if use_df2gsprad:
            if clean is True:
                self.delete_cells(spreadsheet_key, worksheet_name)
            return d2g.upload(df, spreadsheet_key, worksheet_name, credentials=self.creds, row_names=False,
                              value_input_option=value_input_option, clean=clean)

        else:
            if clean is True:
                self.delete_cells(spreadsheet_key, worksheet_name)
            x = Spread(spreadsheet_key, worksheet_name, creds=self.creds, create_sheet=True)
            return x.df_to_sheet(df, index=False, sheet=worksheet_name, replace=clean)

    @retry((requests.exceptions.ReadTimeout, gspread.exceptions.APIError), tries=3, delay=2)
    def pd_read_gsheet(self, spreadsheet_key, worksheet_name):
        """
        :param spreadsheet_key: id for Spreadsheet taken from URL
        :param worksheet_name: name as visibile in worksheet
        :return: pandas DataFrame from worksheet
        """
        try:
            return g2d.download(gfile=spreadsheet_key, wks_name=worksheet_name, col_names=True, credentials=self.creds)
        except JSONDecodeError as e:
            return self.pull_sheet_data(spreadsheet_key, cell_range=worksheet_name)


    def pull_sheet_data(self, spreadsheet_key, cell_range=None):
        sheet = self.service.spreadsheets()
        result = sheet.values().get(
            spreadsheetId=spreadsheet_key,
            range=cell_range).execute()
        values = result.get('values', [])

        if not values:
            print('No data found.')
        else:
            rows = sheet.values().get(spreadsheetId=spreadsheet_key,
                                      range=cell_range).execute()
            data = rows.get('values')
            print("COMPLETE: Data copied")
            import pandas as pd
            return pd.DataFrame(data[1:], columns=data[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Matteo Senardi
# @Date:   2021-04-04 10:57:06
# @Email:  pualien@gmail.com
# @Last modified by:   pualien
# @Last Modified time: 2021-04-04 10:57:06


from httplib2 import Http

from apiclient import discovery, errors

from .utils import logr


def get_file_id(credentials, gfile, write_access=False):
    """
        Get file ID by provided path. If file does not exist and
        `write_access` is true, it will create whole path for you.

        :param credentials: provide own credentials
        :param gfile: path to Google Spreadsheet
        :param write_access: allows to create full path if file does not exist
        :type credentials: class 'oauth2client.client.OAuth2Credentials'
        :type gfile: str
        :type write_access: boolean
        :returns: file ID
        :rtype: str

        :Example:

            >>> from df2gspread_ext.gfiles import get_file_id
            >>> from df2gspread_ext.utils import get_credentials
            >>> gfile = '/some/folder/with/file'
            >>> credentials = get_credentials()
            >>> get_file_id(credentials=credentials, gfile=gfile, write_access=True)
            u'78asbcsSND8sdSACNsa7ggcasca8shscaSACVD'
    """
    # auth for apiclient
    http = credentials.authorize(Http())
    service = discovery.build('drive', 'v3', http=http, cache_discovery=False)

    file_id = service.files().get(fileId='root', fields='id').execute().get('id')

    # folder/folder/folder/spreadsheet
    pathway = gfile.strip('/').split('/')

    for idx, name in enumerate(pathway):
        files = service.files().list(
            q="name = '{}' and trashed = false and '{}' in parents".format(name, file_id)).execute()['files']
        if len(files) > 0:
            # Why do you ever need to use several folders with the same name?!
            file_id = files[0].get('id')
        elif write_access == True:
            body = {
                'mimeType': 'application/vnd.google-apps.' + ('spreadsheet' if idx == len(pathway)-1 else 'folder'),
                'name': name,
                'parents': [file_id]
            }
            file_id = service.files().create(body=body, fields='id').execute().get('id')
        else:
            return None
    return file_id


def get_worksheet(gc, gfile_id, wks_name, write_access=False, new_sheet_dimensions=(1000, 100)):
    """DOCS..."""

    spsh = gc.open_by_key(gfile_id)

    # if worksheet name is not provided , take first worksheet
    if wks_name is None:
        wks = spsh.sheet1
    # if worksheet name provided and exist in given spreadsheet
    else:
        try:
            wks = spsh.worksheet(wks_name)
        except:
            #rows, cols = new_sheet_dimensions
            wks = spsh.add_worksheet(
                wks_name, *new_sheet_dimensions) if write_access == True else None

    return wks


def delete_file(credentials, file_id):
    """DOCS..."""
    try:
        http = credentials.authorize(Http())
        service = discovery.build(
            'drive', 'v3', http=http, cache_discovery=False)
        service.files().delete(fileId=file_id).execute()
    except errors.HttpError as e:
        logr.error(e)
        raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Matteo Senardi
# @Date:   2021-04-04 10:57:06
# @Email:  pualien@gmail.com
# @Last modified by:   pualien
# @Last Modified time: 2021-04-04 10:57:06


import logging
import os
import json

from google.oauth2 import service_account
from oauth2client import file, client, tools

# Load logging before anything else
logr = logging.getLogger('df2gspread')

''' Load the file with credentials '''
CLIENT_SECRET_FILE = os.path.join( os.path.expanduser("~"), ".gdrive_private")

DEFAULT_TOKEN = os.path.join( os.path.expanduser("~"), ".oauth", "drive.json")

# FIXME: clarify scopes
SCOPES = ('https://www.googleapis.com/auth/drive.metadata.readonly '
          'https://www.googleapis.com/auth/drive '
          'https://spreadsheets.google.com/feeds '
          'https://docs.google.com/feeds')


def get_credentials(credentials=None, client_secret_file=CLIENT_SECRET_FILE, refresh_token=None):
    """Consistently returns valid credentials object.

    See Also:
        https://developers.google.com/drive/web/quickstart/python

    Args:
        client_secret_file (str): path to client secrets file, defaults to .gdrive_private
        refresh_token (str): path to a user provided refresh token that is already
            pre-authenticated
        credentials (`~oauth2client.client.OAuth2Credentials`, optional): handle direct
            input of credentials, which will check credentials for valid type and
            return them

    Returns:
        `~oauth2client.client.OAuth2Credentials`: google credentials object

    """

    # if the utility was provided credentials just return those
    if credentials:
        if _is_valid_credentials(credentials):
            # auth for gspread
            return credentials
        else:
            print("Invalid credentials supplied. Will generate from default token.")

    token = refresh_token or DEFAULT_TOKEN
    dir_name = os.path.dirname(DEFAULT_TOKEN)
    try:
        os.makedirs(dir_name)
    except OSError:
        if not os.path.isdir(dir_name):
            raise
    store = file.Storage(token)
    credentials = store.get()

    try:
        import argparse
        flags = argparse.ArgumentParser(
            parents=[tools.argparser]).parse_known_args()[0]
    except ImportError:
        flags = None
        logr.error(
            'Unable to parse oauth2client args; `pip install argparse`')

    if not credentials or credentials.invalid:

        flow = client.flow_from_clientsecrets(
            client_secret_file, SCOPES)
        flow.redirect_uri = client.OOB_CALLBACK_URN
        if flags:
            credentials = tools.run_flow(flow, store, flags)
        else:  # Needed only for compatability with Python 2.6
            credentials = tools.run(flow, store)
        logr.info('Storing credentials to ' + DEFAULT_TOKEN)

    return credentials


def _is_valid_credentials(credentials):
    return isinstance(credentials, client.OAuth2Credentials) or isinstance(credentials, service_account.Credentials)


def create_service_credentials(private_key_file=None, client_email=None,
                               client_secret_file=CLIENT_SECRET_FILE):
    """Create credentials from service account information.

    See Also:
        https://developers.google.com/api-client-library/python/auth/service-accounts

    Args:
        client_secret_file (str): path to json file with just the client_email when
            providing the `private_key_file` separately, or this file can have both the
            `client_email` and `private_key` contained in it. Defaults to .gdrive_private
        client_email (str): service email account
        private_key_file (str): path to the p12 private key, defaults to same name of file
            used for regular authentication

    Returns:
        `~oauth2client.client.OAuth2Credentials`: google credentials object

    """
    if private_key_file is not None:
        with open(os.path.expanduser(private_key_file)) as f:
            private_key = f.read()
    else:
        private_key = None

    if client_email is None:
        with open(os.path.expanduser(client_secret_file)) as client_file:
            client_data = json.load(client_file)

            if 'installed' in client_data:

                # handle regular json format where key is separate
                client_email = client_data['installed']['client_id']
                if private_key is None:
                    raise RuntimeError('You must have the private key file \
                                       with the regular json file. Try creating a new \
                                       public/private key pair and downloading as json.')
            else:
                # handle newer case where json file has everything in it
                client_email = client_data['client_email']
                private_key = client_data['private_key']

    if client_email is None or private_key is None:
        raise RuntimeError(
            'Client email and/or private key not provided by inputs.')

    credentials = client.SignedJwtAssertionCredentials(
        client_email, private_key, SCOPES)

    return credentials



import google.auth
import pandas as pd
from google.analytics.data_v1beta import BetaAnalyticsDataClient
from google.analytics.data_v1beta.types import DateRange
from google.analytics.data_v1beta.types import Dimension
from google.analytics.data_v1beta.types import Filter
from google.analytics.data_v1beta.types import FilterExpression
from google.analytics.data_v1beta.types import FilterExpressionList
from google.analytics.data_v1beta.types import Metric
from google.analytics.data_v1beta.types import RunReportRequest
from google.oauth2 import service_account

from gcloud_connectors.logger import EmptyLogger

SCOPES = ['https://www.googleapis.com/auth/analytics']


class GAnalytics4Connector:
    def __init__(self, confs_path=None, auth_type='service_account', json_keyfile_dict=None, logger=None):
        self.confs_path = confs_path
        self.json_keyfile_dict = json_keyfile_dict
        self.auth_type = auth_type
        self.logger = logger if logger is not None else EmptyLogger()

        if self.confs_path is not None or self.json_keyfile_dict is not None:
            if self.confs_path is not None:
                self.creds = service_account.Credentials.from_service_account_file(
                    self.confs_path,
                )
            elif self.json_keyfile_dict is not None:
                self.creds = service_account.Credentials.from_service_account_info(
                    self.json_keyfile_dict
                )

        else:
            self.creds, project = google.auth.default()

        self.logger = logger if logger is not None else EmptyLogger()
        self.service = BetaAnalyticsDataClient(credentials=self.creds)

    @staticmethod
    def get_base_report_params(property_id, metrics_dict_values, dimensions_dict_values, date_range, limit, offset,
                               keep_empty_rows, return_property_quota):
        return {
            "property": "properties/{property_id}".format(property_id=property_id),
            "dimensions": dimensions_dict_values,
            "metrics": metrics_dict_values,
            "date_ranges": date_range,
            "limit": limit,
            "offset": offset,
            "keep_empty_rows": keep_empty_rows,
            "return_property_quota": return_property_quota
        }

    @staticmethod
    def get_filter_type(filter_type_string):
        filter_type = Filter.StringFilter  # default
        if filter_type_string == "string_filter":
            filter_type = Filter.StringFilter
        elif filter_type_string == "in_list_filter":
            filter_type = Filter.InListFilter
        elif filter_type_string == "numeric_filter":
            filter_type = Filter.NumericFilter
        elif filter_type_string == "between_filter":
            filter_type = Filter.BetweenFilter
        return filter_type

    def get_filters(self, filters):
        """
        filters = [{"field_name": "browser", "type": "string_filter", "value": "chrome", "match_type": 1,
        "is_not": True}]
        """
        filter_dict_values = []
        for filter_object in filters:
            filter_type = self.get_filter_type(filter_object["type"])
            if filter_object.get("match_type"):
                filter_applied = filter_type(
                    {"value": filter_object["value"], "match_type": filter_object["match_type"]}
                )
            else:
                if filter_object["type"] == 'in_list_filter':
                    filter_applied = filter_type({"values": filter_object["values"]})
                else:
                    filter_applied = filter_type({"value": filter_object["value"]})

            # TODO: wait to be fixed combining not_expressions inside FilterExpressionList. In march 2022 leads to
            #  TypeError: Parameter to MergeFrom() must be instance of same class: expected FilterExpression
            #  got Filter. for field FilterExpression.not_expression
            if filter_object.get("is_not"):
                filter_dict_values.append(
                    FilterExpression({
                        "not_expression": Filter({
                            "field_name": filter_object["field_name"],
                            filter_object["type"]: filter_applied

                        }
                        )
                    })
                )
            else:
                filter_dict_values.append(
                    FilterExpression({
                        "filter": Filter({
                            "field_name": filter_object["field_name"],
                            filter_object["type"]: filter_applied

                        }
                        )
                    })
                )
        return filter_dict_values

    @staticmethod
    def get_df_from_response(response, dimensions, metrics):
        output = []
        for row in response.rows:
            current_row = {}
            for dim in dimensions:
                current_row[dim] = row.dimension_values[dimensions.index(dim)].value
            for metr in metrics:
                current_row[metr] = row.metric_values[metrics.index(metr)].value
            output.append(current_row)
        return pd.DataFrame(output)

    def pd_get_report(self, property_id, start_date, end_date, metrics, dimensions, dimension_filters=None,
                      metrics_filters=None, offset=0, limit=100000,
                      keep_empty_rows=False, return_property_quota=False, downloaded_totals=0, cast_date_column=None):
        """Runs a complex report on a Google Analytics 4 property."""
        dimensions_dict_values = [Dimension({"name": x}) for x in dimensions]
        metrics_dict_values = [Metric({"name": x}) for x in metrics]
        date_range = [DateRange({"start_date": start_date, "end_date": end_date})]

        if dimension_filters is None:
            dimension_filters = []
        if metrics_filters is None:
            metrics_filters = []

        dimension_filter_dict_values = self.get_filters(dimension_filters)
        metric_filter_dict_values = self.get_filters(metrics_filters)

        report_params = self.get_base_report_params(property_id=property_id, metrics_dict_values=metrics_dict_values,
                                                    dimensions_dict_values=dimensions_dict_values,
                                                    date_range=date_range, limit=limit, offset=offset,
                                                    keep_empty_rows=keep_empty_rows,
                                                    return_property_quota=return_property_quota)
        if len(dimension_filter_dict_values):
            report_params["dimension_filter"] = FilterExpression({
                "and_group": FilterExpressionList(
                    {
                        "expressions": dimension_filter_dict_values

                    }
                )
            })
        if len(metric_filter_dict_values):
            report_params["metric_filter"] = FilterExpression({
                "and_group": FilterExpressionList(
                    {
                        "expressions": metric_filter_dict_values
                    }
                )
            })

        request = RunReportRequest(report_params)

        response = self.service.run_report(request)

        df = self.get_df_from_response(response, dimensions, metrics)

        # if missing pages for all rows in response.row_count
        if len(response.rows) < response.row_count and downloaded_totals < response.row_count:
            downloaded_totals = len(response.rows) + downloaded_totals
            print(
                "Recursive iteration page {current_page}/{pages} for records {records}/{totals}".format(
                    records=offset + limit if offset + limit < response.row_count else response.row_count,
                    totals=response.row_count,
                    current_page=round(downloaded_totals / limit),
                    pages=round(response.row_count / limit)
                ))
            df = df.append(
                self.pd_get_report(property_id=property_id, start_date=start_date, end_date=end_date, metrics=metrics,
                                   dimensions=dimensions, dimension_filters=dimension_filters,
                                   metrics_filters=metrics_filters, offset=offset + limit, limit=limit,
                                   keep_empty_rows=keep_empty_rows, return_property_quota=return_property_quota,
                                   downloaded_totals=downloaded_totals))
        if cast_date_column:
            if 'date' in df.columns:
                df['date'] = pd.to_datetime(df['date'], format="%Y%m%d").dt.date
        return df



import logging


class Logger:
    def __init__(self, logger_name):
        logger = logging.getLogger(logger_name)

        logger.setLevel(logging.INFO)
        # create console handler with a higher log level
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        # create formatter and add it to the handlers
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        # add the handlers to logger
        logger.addHandler(ch)
        self.logger = logger

    def get_logger(self):
        return self.logger


class EmptyLogger:
    @staticmethod
    def info(msg):
        pass

    @staticmethod
    def warn(msg):
        pass

    @staticmethod
    def warning(msg):
        pass

    @staticmethod
    def debug(msg):
        pass

    @staticmethod
    def critical(msg):
        pass

    @staticmethod
    def error(msg):
        pass
